this.name        = "telescope_fps_monitor";
this.author      = "cag";
this.copyright   = "2017 cag";
this.licence     = "CC BY-NC-SA 4.0";
this.description = "Monitor and report frames/sec, and optionally, various other stats, plus callback support.";
this.version     = "1.2";

/* global addFrameCallback, isValidFrameCallback, log, player, removeFrameCallback, Uint8Array
 */
 
/// stripped down version of fps_monitor for running with Telescope 

this._fps_monitor_closure = (function _fps_monitor_closure() { 
    "use strict";
	// closure is self-initiating, doing so when it loads, so there no need to call it

	var that = _fps_monitor_closure;
    
    /*  user settings  */
    
    var OXP_NAME = '<<your oxp>>'; // required! used as 1st parm in call to log(); can also be set 
                                   //   if you call _init_fps_monitor( '<<your oxp>>' )
    // reporting
    var N_MINUTES = 1;             // minutes, frequency of reporting on last minute's stats (default 1, 0 to turn off)
    var SHORT_TERM = 15;           // minutes, > 1, time interval for short term report (default 15, 0 to turn off)
    var LONG_TERM = 60;            // minutes, > 1, time interval for long term report (default 60, 0 to turn off)
    var FPS_ONLY = false;          // stores and reports only fps value
    var FILE_LOG = true;           // write to the log file results for 3 time periods just mentioned
    var CONSOLE_LOG = false;       // echo the reports to in-game console
    var CONSOLE_TIME = 5;          // seconds, 1-10, duration of in-game console message (default 5)

    // calculations
    var TRUNC_LOW = 10;            // seconds, > 0, for discarding low outliers before calculating fps (default 10, 0 to turn off)
    var TRUNC_HIGH = 5;            // seconds, > 0, for discarding high outliers before calculating fps (default 5, 0 to turn off)
    var HARMONIC_MEAN = false;     // when true, calc harmonic rather than arithmetic mean, for less sensitivity to outliers

    // these next 5 booleans turn on extra values for reports and return values for data fetch functions
    // - computed every minute together w/ fps; relative order is always:  median, mode, mean, high, low
    
    var MEDIAN = true;             // fps rate that has equal # of values before/after it in the minute
                                   // - averaged over SHORT_TERM and LONG_TERM
    var MODE = false;              // fps rate with the most hits in the minute
                                   // - averaged over SHORT_TERM and LONG_TERM
    var MEAN = false;              // arithmetic mean (simple average), not very useful, included for completeness
                                   // - averaged over SHORT_TERM and LONG_TERM
    var HIGH = true;               // highest frame count in a second in the term
    var LOW = true;                // lowest frame count in a second in the term

    /*  user functions  */
    
    function _setup_fps_report( minutes, shortterm, longterm, filelog, console, duration ) { // set reporting variables
        // set values of N_MINUTES, SHORT_TERM, LONG_TERM, FILE_LOG, CONSOLE_LOG, CONSOLE_TIME
        //               int,       int,        int,       bool,     bool,        int
		N_MINUTES = get_term( minutes, 1, 1 );
		SHORT_TERM = get_term( shortterm, 1, 15 );
		LONG_TERM = get_term( longterm, 1, 60 );
		if( filelog !== undefined ) FILE_LOG = !!filelog;
		if( console !== undefined ) CONSOLE_LOG = !!console;
		if( N_MINUTES === 0 && SHORT_TERM === 0 && LONG_TERM === 0 ) {
			FILE_LOG = false;
			CONSOLE_LOG = false;
		}
		if( duration ) {						// cannot be set to 0
			let dur = get_term( duration, 1, 3 );
			CONSOLE_TIME = dur > 10 ? 10 : dur;	// as per the wiki
		}
	}
	function _setup_fps_calc( cut_low, cut_high, harmonic, fps_only, median, mode, mean, high, low ) { // set calculation variables
        // set values of TRUNC_LOW, TRUNC_HIGH, HARMONIC_MEAN, FPS_ONLY, MEDIAN, MODE, MEAN, HIGH, LOW
        //               int,       int,        bool,          bool,   	 bool,   bool, bool, bool, bool
		TRUNC_LOW = get_term( cut_low, 1, 5 );
		TRUNC_HIGH = get_term( cut_high, 1, 5 );
		if( harmonic !== undefined )HARMONIC_MEAN = !!harmonic;
		if( fps_only ) FPS_ONLY = !!fps_only;
		if( FPS_ONLY ) MEDIAN = MODE = MEAN = HIGH = LOW = false;
		if( median !== undefined )	MEDIAN = !!median;
		if( mode !== undefined )	MODE = !!mode;
		if( mean !== undefined )	MEAN = !!mean;
		if( high !== undefined )	HIGH = !!high;
		if( low !== undefined )		LOW = !!low;
		if( FPS_ONLY && ( MEDIAN || MODE || MEAN || HIGH || LOW ) ) {
			let num = (MEDIAN ? 1 : 0 ) + (MODE ? 1 : 0 ) + (MEAN ? 1 : 0 )
						+ (HIGH ? 1 : 0 ) + (LOW ? 1 : 0 );
			let count = num;
			log('fps_monitor', '_setup_fps_calc, warning: fps_only setting (true) conflicts with ' 
					+ (MEDIAN ? 'median' + (num-- > 1 ? ', ' : '') : '') 
					+ (MODE ? 'mode'     + (num-- > 1 ? ', ' : '') : '') 
					+ (MEAN ? 'mean'     + (num-- > 1 ? ', ' : '') : '')
					+ (HIGH ? 'high'     + (num-- > 1 ? ', ' : '') : '') 
					+ (LOW ? 'low'       + (num-- > 1 ? ', ' : '') : '') 
					+ '. '+(count>1 ? 'These values are' : 'This value is') +' NOT being stored.' ); 
		}
		if( FPS_ONLY ) MEDIAN = MODE = MEAN = HIGH = LOW = false;
	}
	function _init_fps_monitor( oxp_name, paused, no_fcb ) {// str, [bool, bool]; one-time setup; sets 'oxp_name' used when calling log(OXP_NAME,"...")
															//  - 'paused' prevents it from starting until you call _turn_on_fps_monitor()
															//  - 'no_fcb' prevents it adding a frame callback, so you either call _turn_on_fps_monitor()
															//	  later or one of your fcb's calls _fps_monitor() and passes along 'delta'
		if( oxp_name )
			OXP_NAME = typeof oxp_name === 'string'	// or, you can just edit this file above
							? oxp_name
							: oxp_name.toString();
        else
            OXP_NAME = 'FPS_MONITOR';
		if( isValidFrameCallback( FCB_handle ) ) 
			removeFrameCallback(  FCB_handle );
		if( !no_fcb && !isValidFrameCallback( FCB_handle ) )
			FCB_handle = addFrameCallback( _fps_monitor.bind( that ) );
			// FCB_handle = addFrameCallback( _fps_monitor.bind( this ) );
		if( paused !== undefined )
			is_running = !paused;
		else
			is_running = true;
	}
	function _turn_on_fps_monitor( no_fcb, stutter, thresh, wipe ) {// [bool[,bool[,float[,bool]]]]; resume fps monitoring; set 'no_fcb' to true to prevent adding a frame callback (see above)
																// - set 'stutter' to true to enable stutter checking (NB: this will force LOW to be true!)
																// - 'thresh' is a number > 0 & < 1 that defines what a stutter is, default 0.8.
																//   So if you encounter a frame count that's less than thresh * fps, a report is written, at the end of that minute,
																//   to your log file. (w/ default 0.8, any second w/ < 80% of mean frames gets reported)
																// - 'wipe' set true wipes all data arrays for a clean start
		if( !no_fcb && !isValidFrameCallback( FCB_handle ) )
			FCB_handle = addFrameCallback( _fps_monitor.bind( that ) );
			// FCB_handle = addFrameCallback( _fps_monitor.bind( this ) );
		_reset_fps_monitor( 0, wipe );
		is_running = true; 
		if( stutter === undefined ) return;
		chk_stutter = !!stutter;
		if( chk_stutter && !stutter_mon && !LOW ) 
			set_calc_vars();					// need to adjust for required LOW values
		stutter_mon = chk_stutter;
		if( thresh === undefined ) return;
		if( thresh > 0 && thresh < 1 )
			threshold = thresh;
		else
			log( OXP_NAME, 'fps_monitor, _turn_on_fps_monitor, invalid threshold of ' + thresh + ', using default of ' + threshold );
	}
	function _turn_off_fps_monitor( remove ) {	// [bool]; suspend fps monitoring; set 'remove' to true to remove frame callback
		if( remove && isValidFrameCallback( FCB_handle ) ) {
			removeFrameCallback( FCB_handle );
			FCB_handle = null;
		}
		is_running = false; 
	}
	function _set_callback( fn, period ) {	// function, [int]; returns bool; set a callback function to be invoked every 'period' seconds (default is 60)
											// - for performance reasons, callbacks only occur on second boundary. If you want more, search for the string
											//   "per frame resolution" and uncomment the line
		if( typeof fn !== 'function' ) 	return false;
		if( period <= 0 ) 				return false;
		callback = fn;
		callback_period = period > 0 ? period >>> 0 : 60;
		var max = SHORT_TERM > LONG_TERM ? SHORT_TERM : LONG_TERM;
		if( N_MINUTES > max ) max = N_MINUTES;
		max *= 60;
		if( callback_period > max ) {
			callback_period = max;
			log( OXP_NAME, 'fps_monitor, _set_callback, period exceeds your largest time interval, reset to ' + max );
		}
		return true;
	}
	function _clear_callback() {				// clear a callback function set in _set_callback
		callback = null;
		callback_period = 0;
		return true;
	}
	function _realtime_fps() {					// returns int; result of the last second's frame count
												//  - returns -1 in the event of no data (not started)
												//  - not real useful (?), but trivial to implement
		if( frames_per_sec[ 0 ] === 0 			// has not started
			|| !is_running )					// is not running
			return -1;	
		let i = seconds - 1;					// seconds is post-incremented
		if( i < 0 ) i = 59;
		return frames_per_sec[ i ];				// return last frame's count
	}
	function _current_fps( all_stats, result ) {// [bool[,array]]; returns int or array; result(s) of the last minute's measured data
												//  - returns -1 in the event of no data (not started, not running long enough)
												//  - when 'all_stats' is true, returns an array of up to 6 values.  I deliberately left 
												//    this out of the _setup_... functions, so it would be more apparent what return to expect
												//    - array values have relative order: fps, median, mode, mean, high, low
												//  - when 'result' is present, data is copied into the array (ie. not creating new one to return)
												//    - will append to any existing array data, so set its length = 0 if you want to empty it first
		return return_data( minute_data, all_stats, result );
	}
	function _short_term_fps( all_stats, result ) {// [bool[,array]]; returns int or array; result(s) of the last SHORT_TERM's *reported* data
												//  - if you want SHORT_TERM data from period ending when you call, use _get_fps_data()
												//  - otherwise, works the same as _current_fps()
		return return_data( short_data, all_stats, result );
	}
	function _long_term_fps( all_stats, result ) {// [bool[,array]]; returns int or array; result(s) of the last LONG_TERM's *reported* data
												//  - if you want LONG_TERM data from period ending when you call, use _get_fps_data()
												//  - otherwise, works the same as _current_fps()
		return return_data( long_data, all_stats, result );
	}
	function _get_frames( count, result ) {		// [int[,array]]; returns an array integer frame counts; 'count' specifies # of seconds of 
												// data to return; if no count supplied, all are returned, to a max. of 60 (if it's been running 
												// less than a minute, the array's length will reflect that)
												//  - returns -1 in the event of no data (not started, not running long enough)
												//  - when 'result' is present, data is copied into the array (ie. not creating new one to return)
												//    - will append to any existing array data, so set its length = 0 if you want to empty it first
		var len, asking, data, dst, src;
		if( !is_running 						// monitor is not running
			|| frames_per_sec[ 0 ] === 0 )	
			return -1;		
		len = frames_per_sec[ seconds ] === 0 	// buffer not full, strip unused
				? seconds : 60;
		asking = 60;							// max. # to return
		if( count !== undefined )
			asking = typeof count === 'number' 
						? round( count )
						: parseInt( count, 10 );
		dst = asking < len ? asking : len;		// index for destination
		asking = dst;
		if( result && Array.isArray(result) ) { 
			data = result;						// copy data into supplied array
			dst += result.length;
		} else {
			data = new Array( dst );
		}
		src = seconds;							// index for source
		while( asking-- ) {
			if( src <= 0 ) src = 60;
			data[ --dst ] = frames_per_sec[ --src ];
		}
		return data;
	}
	function _get_fps_data( all_stats, count, result ) {// [bool[,int[,array]]]; returns an array of integers of all it's collected for further processing
												//  - returns -1 in the event of no data (not started, not running long enough)
												//  - by default, returns an array of fps values
												//    - set 'all_stats' true for data to be an array of: fps, median, ..., depending on what
												//      flags you set in _setup_fps_calc() (or by editing the script)
												//      NB: not an array of [ fps, median, ... ] entries, just a stream of int's
												//  - length of return array will be max( SHORT_TERM, LONG_TERM ) if !all_stats
												//    - if all_stats is true, each min. will have 1 - 6 int's, depending on flags set previously 
												//  - count is some # of minutes <=  max( SHORT_TERM, LONG_TERM )
												//  NB: _short_term_fps() returns data from short term report, as such, only changes every SHORT_TERM minutes.
												//      If you want data from the previous SHORT_TERM minutes, call this fn w/ count = SHORT_TERM. Ditto for long term.
												//  - when 'result' is present, data is copied into the array (ie. not creating new one to return)
												//    - will append to any existing array data, so set its length = 0 if you want to empty it first
		var len, asking, data, fps_only, maxm, dst, src;
		if( !is_running || !fps_stats			// monitor is not running
			|| fps_stats[ 0 ] === 0 )			// too soon, have not measured anything
			return -1;		
		fps_only = all_stats !== undefined
					? !all_stats : true;
		if( fps_stats[ index ] === 0 ) {		// buffer not full, strip unused
			maxm = asking = minutes;			// default for when !count
			len = index;
		} else {
			maxm = asking = max_minutes;		// default for when !count
			len = array_size;					// max. fps_stats index
		}
		if( count ) {
			asking = typeof count === 'number' 
						? round( count )
						: parseInt( count, 10 );
		}
		if( asking > maxm ) asking = maxm;
		if( fps_only ) {
			dst = asking;						// index for destination			
		} else {
			let dsize = asking * DATA_SIZE;			
			dst = dsize < len ? dsize : len;	// index for destination
		}
		if( result && Array.isArray(result) ) {
			data = result;						// copy data into supplied array
			dst += result.length;
		} else {
			data = new Array( dst );
		}
		src = index;							// index for source
		while( asking-- ) {
			if( src <= 0 ) src = array_size;
			if( fps_only ) {
				src -= DATA_SIZE - 1;			// discard everything but fps
			} else {
				if( LOW )    data[ --dst ] = fps_stats[ --src ];
				if( HIGH )   data[ --dst ] = fps_stats[ --src ];
				if( MEAN )   data[ --dst ] = fps_stats[ --src ];
				if( MODE  )  data[ --dst ] = fps_stats[ --src ];
				if( MEDIAN ) data[ --dst ] = fps_stats[ --src ];
			}
			data[ --dst ] = fps_stats[ --src ];	// copy fps
		}
		return data;
	}

	/*  internal variables - you break it, you own it!  */

	// function references
	var round = Math.round;
	var callback = null;						// function to be called at end of each callback_period minutes
	// local variables
	var chk_stutter = false;					// user toggle for stutter monitor
	var stutter_mon = false;					// internal state for stutter monitor
	var threshold = 0.8;						// amount below average fps that defines what's a stutter (eg. < 80%)
	var stutters = [];							// array (re-used) to store stutters while reporting
	var callback_period = 0;					// # of seconds between each callback
	var game_paused = false;					// flag so when games resumes, can continue accuratly
	var FCB_handle = null;						// reference for frame callback
	var is_running = false;						// on - off switch available via _turn_on_fps_monitor() & _turn_off_fps_monitor()
	var frames = 0;								// # of frames
	var seconds = 0;							// # of seconds, index into frames_per_sec array
	var minutes = 0;							// # of minutes; not an index, so keeps incrementing (unles _reset_fps_monitor( .., true ))
	var start_delta = -1;						// delta value @ start of ea. sec. (can be any value, depending on when 1st fcb called)
	var running_total = 0;						// running total of delta values passed to frame callback
	var index = 0;								// index into fps_stats, max is array_size
	var frames_per_sec = new Uint8Array( 60 );	// typed array used for storing frame counts, wraps every minute
	var counted = new Uint8Array( 256 ); 		// internal working area for counting_sort; frame counts limited to 254!
	var counted_most, counted_mid;				// when sorting, save most popular -> mode, middle -> median

	var DIVISOR = 60 - TRUNC_HIGH - TRUNC_LOW;	// divisor for (trucated) mean or dividend for harmonic mean
	var DATA_SIZE, array_size, max_minutes, fps_stats, long_data, short_data, minute_data, mid_data, frames_sorted;
	// NB: data store relative order: fps, median, mode, mean, high, low
	//     - some may be missing due to user settings
	function set_calc_vars() {
		DIVISOR = 60 - TRUNC_HIGH - TRUNC_LOW;
		max_minutes = LONG_TERM > SHORT_TERM	// the larger of the two
				   ? LONG_TERM : SHORT_TERM;
		var old_data_size = DATA_SIZE;
		if( FPS_ONLY ) {
			frames_sorted = frames_per_sec;		// sort not called, pts to frames_per_sec for code following sort
			DATA_SIZE = (stutter_mon || chk_stutter) ? 2 : 1;
		} else {
			frames_sorted = new Uint8Array( 60 );// typed array used for calc'g average fps, high & low values, reset every minute
			DATA_SIZE = 1 + (MEDIAN ? 1 : 0 )	// saving 1 data point + upto 5 more
						  + (MODE   ? 1 : 0 )
						  + (MEAN   ? 1 : 0 )
						  + (HIGH   ? 1 : 0 )
						  + (LOW || stutter_mon || chk_stutter ? 1 : 0 );
		}
		if( old_data_size === DATA_SIZE ) 	 	// subsequent call but size not changed
			return;	
		fps_stats = null;
		array_size = max_minutes * DATA_SIZE;
		if( array_size > 0 )
			fps_stats = new Uint8Array( array_size );// a 2nd typed array of each minute's stats as an int or run of ints
		minute_data = new Uint8Array( DATA_SIZE );// if both SHORT_TERM & LONG_TERM shut off, still needed for minutes
		short_data = long_data = mid_data = null;
		if( SHORT_TERM === 0 && LONG_TERM === 0 ) {
			return;
		} else if( SHORT_TERM === 0 ) {
			long_data = new Uint8Array( DATA_SIZE );// save latest data for get fns
		} else if( LONG_TERM === 0 ) { 
			short_data = new Uint8Array( DATA_SIZE );
		} else {
			short_data = new Uint8Array( DATA_SIZE );
			long_data = new Uint8Array( DATA_SIZE );	// save latest data for get fns
			mid_data = new Uint8Array( DATA_SIZE );	// internal for case when long & short reports coincide
		}			
	}
	
	var rpt_begin = '*** FPS report ***: over the last ';
	var minutes_str = [ 'minute', '1 min'];
	var short_term_str, long_term_str;
	function set_time_strs() {
		short_term_str = [ (SHORT_TERM % 15 === 0
							? (SHORT_TERM / 15) + '/4 hour'
								+ (SHORT_TERM > 15 ? 's.' : '.') 
							: SHORT_TERM + ' minutes'),
							SHORT_TERM + ' min' ];
		long_term_str = [ (LONG_TERM  % 60 === 0 
							? (LONG_TERM / 60) +' hour' 
								+ (LONG_TERM > 60 ? 's.' : '.') 
							: LONG_TERM + ' minutes'),
							LONG_TERM + ' min' ];
	}

	/*  internal functions - any modification will void you warrantee  */
	
	function get_term( term, minimum, defawlt ) {
		var num;
		if( !term ) return 0;					// no parms shuts it down
		num = typeof term === 'number' 
				? round( term ) : parseInt( term, 10 );
		num = num >= minimum ? num : defawlt;
		return num;
	}
	function return_data( data, all_stats, result ) {
		if( !data || data[ 0 ] === 0  			// too soon, have not measured 1st minute/term
				  || !is_running )				// is not running
			return -1;	
		var fps_only, rtn, len, i;
		fps_only = all_stats === undefined
						? true : !all_stats;
		if( !all_stats && !result )
			return data[ 0 ];					// return fps
		
		if( result && Array.isArray(result) ) {
			i = result.length;
			rtn = result;
		} else {
			i = 0;
			rtn = [];
		}					
		len = fps_only ? 1 : DATA_SIZE;
		while( len-- ) {
			rtn[i] = data[i];					// copy data into (supplied) array		
			i++;
		}
		return rtn;								// return [ fps, ... ]
	}
	function wipe_data( list ) {				// for typed Arrays, setting length = 0 doesn't clear, .fill() not supported
		if( list ) for( var i = list.length-1; i >= 0; i-- ) list[i] = 0;
	}
	function _reset_fps_monitor( delta, restart ) {	// [float[,bool]]: reset the monitor; internal function, useful for limiting monitoring (eg. dockside vs in flight)
													//  - not strictly required, as invoking the closure initializes these
													//  - if using this, it's not necessary to supply 'delta' (counting will start on next frame)
													//  - 'restart' sets minutes to 0 & wipes existing data; for calls w/o 'delta', use _reset_fps_monitor( 0, true )
		if( restart === true ) {
			wipe_data( minute_data );
			wipe_data( fps_stats );
			wipe_data( short_data );
			wipe_data( long_data );
			wipe_data( mid_data );
			minutes = 0;
		}
		start_delta = 	!delta ? -1 : running_total;
		running_total = !delta ? 0  : running_total + delta;
		frames =		!delta ? 0  : 1;							
		seconds = 0;							// indices must be set to 0	every minute
		index = 0;								//    "                     every 'array_size'
	}
	function rpt_to_console( fps, median, mode, mean, high, low, time_str ) {
		let msg = 'FPS: ' + fps;
		if( !FPS_ONLY )  {
			if( MEDIAN ) msg += ', median: ' + median;
			if( MODE )   msg += ', mode: ' + mode;
			if( MEAN )   msg += ', mean: ' + mean;
			if( HIGH )   msg += ', hi: ' + high;
			if( LOW )    msg += ', lo: ' + low;
		}
		msg += ' (' + time_str[ 1 ] + ')';
		player.consoleMessage( msg, CONSOLE_TIME );
	}
	function rpt_to_log( fps, median, mode, mean, high, low, time_str ) {
		var type = '';
		if( HARMONIC_MEAN ) type = 'harmonic';
		if( TRUNC_HIGH === 25 && TRUNC_LOW === 25 ) 
			type += 'interquartile';
		else if( TRUNC_HIGH || TRUNC_LOW ) type = 'truncated' + (type ? ' ' + type: '');
		let report = rpt_begin + time_str[ 0 ] + ', '+type+' mean fps was '+ fps;
		if( !FPS_ONLY ) {
			if( MEDIAN ) report += ', median: ' + median;
			if( MODE )   report += ', mode: ' + mode;
			if( MEAN )   report += ', mean: ' + mean;
			if( HIGH )   report += ', high: ' + high;
			if( LOW )    report += ', low: ' + low;
		}
		log( OXP_NAME, report );
	}
	function rpt_results( data, time_str ) {
		var fps, median, mode, mean, high, low;
		var i = 0;
		fps = data[ i++ ];
		median = mode = mean = high = low = 0;
		if( MEDIAN ) median = data[ i++ ];
		if( MODE )   mode =   data[ i++ ];
		if( MEAN )   mean =   data[ i++ ];
		if( HIGH )   high =   data[ i++ ];
		if( LOW )    low =    data[ i++ ];
		if( FILE_LOG )
			rpt_to_log( fps, median, mode, mean, high, low, time_str );
		if( CONSOLE_LOG ) 
			rpt_to_console( fps, median, mode, mean, high, low, time_str );
	}
	function check_stutter() {
		var fps = minute_data[ 0 ];
		if( !fps ) return;	// not ready
		var low = minute_data[ DATA_SIZE - 1 ];
		var thresh = fps * threshold;
		if( low >= thresh ) return; // ignore any diff above threshold
		var ref = '', all = '[ ', rpt = '';
		var hic_len = 0, run = 0, slot, found = false;
		var i, frames;
		for( i = 0; i < 60; i++ ) { // always called at end of a full minute
			frames = frames_per_sec[ i ];
			slot = i + 1;
			if( i < 15 ) ref += slot > 9 ? '   ' + slot : '    ' + slot;
			all += (frames < 100 ? ' ' : '') + frames + (i === 59 ? ' ]' : (i < 59 && slot % 15 === 0) ? ',\n  ' : ', ');
			if( frames < thresh ) {
				stutters.push( frames );
				found = true;
				continue;
			}
			if( found ) { 
				if( run > 0 ) rpt += ' ..[' + run + ' s].. ';
				hic_len = stutters.length;
				while( hic_len-- ) 
					rpt += stutters.shift() + (hic_len > 0 ? ', ' : ' ');
				stutters.length = hic_len = run = 0;
				found = false;
			}
			run++;
		}
		rpt += run > 0 ? '..[' + run + ' s]..' : '';
		if( found ) { 
			hic_len = stutters.length;
			while( hic_len-- ) 
				rpt += stutters.shift() + (hic_len > 0 ? ', ' : ' ');
			stutters.length = 0;
		}
		ref += '\n';
		rpt = 'check_stutter, w/ threshold = ' + thresh.toFixed(1) + ' frames:  ->  ' + rpt;
		log( OXP_NAME, rpt + '\nfps = ' + fps + ', low = ' + low + ', last 60 seconds:\n' + ref + all );
	}
/*
*** FPS report ***: over the last minute, harmonic mean fps was 140, median: 139, high: 205, low: 130

check_stutter, w/ threshold = 133.0 frames:  ->   ..[39 s].. 130  ..[7 s].. 132 ..[12 s]..
fps = 140, low = 130, last 60 seconds:
    1    2    3    4    5    6    7    8    9   10   11   12   13   14   15
[ 146, 137, 140, 134, 137, 140, 135, 137, 139, 137, 145, 139, 146, 138, 137,
  133, 135, 145, 138, 139, 139, 143, 142, 141, 141, 139, 145, 140, 138, 142,
  133, 139, 134, 138, 149, 145, 138, 139, 136, 130, 146, 136, 138, 139, 138,
  137, 144, 132, 140, 142, 137, 137, 135, 134, 133, 137, 137, 152, 205, 205 ]
*/
	function counting_sort() {
		var i, cs, max = 0, hh, ll, diff;
		cs = 255; 
		while( cs-- ) counted[ cs ] = 0;		// clear counts array
		i = 60;
		while( i-- ) {
			cs = frames_per_sec[ i ];			// record count of each value
			counted[ cs ]++;
		}
		cs = 255; i = 60;
		while( cs-- ) {
			let count = counted[ cs ];			// re-build frames_per_sec sorted by counts
			if( MODE && count > max ) { 
				max = count;
				counted_most = cs;
			}
			while( count-- ) {
				frames_sorted[ --i ] = cs;
				if( MEDIAN && i > 28 && i < 31 ) {
					if( i === 30 ) hh = cs;
					else if( i === 29 ) {
						ll = cs;
						diff = hh - ll;
						if( diff % 2 === 0 ) 	// ll + diff/2
							counted_mid = ll + (diff >>> 1);
						else  					// round( ll + diff/2 )
							counted_mid = ll + ((diff + 1) >>> 1);
					}
				}
			}
		}
	}
	function do_callback() {
		if( !callback ) return;
		if( callback_period <= 0 ) return;
		var elapsed = callback_period <= 60 ? seconds : minutes * 60 + seconds;
		if( elapsed % callback_period === 0 ) {
			try {
				callback();
			} catch( err ) {
				log( OXP_NAME, 'callback function encountered an error: ' + err );
				_clear_callback();
				log( OXP_NAME, 'callback function has been cleared.' ); 
			}
		}
	}
	function _fps_monitor( delta ) {
		if( delta === 0 && game_paused ) {		// game paused, subsequent frames
			return;
		} else if( delta === 0 ) {				// game paused, first frame
			game_paused = true;
			return;
		} else if( delta > 0 && game_paused ) {	// game resumed
			game_paused = false;
		}
		var fps, high, low, highest, lowest, short_rpt, long_rpt,
			count, sum, total, i, j, len, mid, mean, mode, median, 
			mean_sum, mode_sum, median_sum, div, data, mid_div, mid_data;
		if( !is_running ) return;
		if( !frames_sorted ) set_calc_vars();	// just-in-time init'n
		if( index >= array_size ) index = 0;	// must be @ top of 'loop' due to various return statements
		if( start_delta === -1 ) {				// 1st time thru
			running_total = start_delta = delta;
			frames = 0;							// start counting next frame
			return;
		}
		running_total += delta;
		frames++;
		if( running_total - start_delta < 1 ) {
///			do_callback();						// remove comment for per frame resolution
			return;
		}										// every second
		frames_per_sec[ seconds++ ] = frames > 254 ? 254 : frames;
		frames = 0;
		start_delta = running_total;
		if( seconds < 60 ) {
			do_callback();
			return;	
		}										// every minute	
		seconds = 0;
		minutes++;
		if( !FPS_ONLY )
			counting_sort();					// req'd for truncated mean, median & mode
		median = mode = mean = high = sum = total = 0;
		low = 254; i = 60;
		while( i-- ) {
			count = frames_sorted[ i ];
			total += count;
			if( i >= TRUNC_LOW && i < (60 - TRUNC_HIGH) ) {
				if( HARMONIC_MEAN ) {
					if( count > 0 )sum += 1/count;
				} else
					sum += count;
			}
			if( HIGH && count > high ) high = count;
			if( (LOW || stutter_mon) && count < low ) low = count;
		}
		fps = HARMONIC_MEAN ? round( DIVISOR / sum )
							: round( sum / DIVISOR );
		i = 0;
		minute_data[ i++ ] = fps;
		if( MEDIAN ) minute_data[ i++ ] = counted_mid;
		if( MODE )   minute_data[ i++ ] = counted_most;
		if( MEAN )   minute_data[ i++ ] = round( total/ 60 );
		if( HIGH )   minute_data[ i++ ] = high;
		if( LOW || stutter_mon ) minute_data[ i++ ] = low;
		if( N_MINUTES > 0 && (N_MINUTES === 1	// not turned off by user
				|| minutes % N_MINUTES === 0) ) // on Nth minute
			rpt_results( minute_data, minutes_str );
		if( stutter_mon ) check_stutter();
		if( array_size === 0 ) {
			do_callback();
			return;								// turned off by user
		}
		i = 0; // store data for short & long term reports
		fps_stats[ index++ ] = minute_data[ i++ ]; // fps
		if( MEDIAN ) fps_stats[ index++ ] = minute_data[ i++ ];
		if( MODE )   fps_stats[ index++ ] = minute_data[ i++ ];
		if( MEAN )   fps_stats[ index++ ] = minute_data[ i++ ];
		if( HIGH )   fps_stats[ index++ ] = minute_data[ i++ ];
		if( LOW || stutter_mon ) fps_stats[ index++ ] = minute_data[ i++ ];
						
		short_rpt = SHORT_TERM > 0 && minutes > 1 
						? minutes % SHORT_TERM : -1;
		long_rpt =  LONG_TERM > 0 && minutes > 1 
						? minutes % LONG_TERM  : -1;
		if( short_rpt !== 0 && long_rpt !== 0 ) {// not ready to report either benchmark
			do_callback();
			return;							
		}
		if( short_rpt === long_rpt ) {			// reporting both
			if( SHORT_TERM < LONG_TERM ) {
				len = j = LONG_TERM * DATA_SIZE;
				data = long_data;
				div = LONG_TERM;
				mid = SHORT_TERM * DATA_SIZE;
				mid_data = short_data;
				mid_div = SHORT_TERM;
			} else {
				len = j = SHORT_TERM * DATA_SIZE;
				data = short_data;
				div = SHORT_TERM;
				mid = LONG_TERM * DATA_SIZE;
				mid_data = long_data;
				mid_div = LONG_TERM;
			}
		} else {								// reporting which one?
			if( long_rpt === 0  ) {
				len = j = LONG_TERM * DATA_SIZE;
				data = long_data;
				div = LONG_TERM;
			} else {
				len = j = SHORT_TERM * DATA_SIZE;
				data = short_data;
				div = SHORT_TERM;
			}
			mid = -1;							// not used for single, act as flag
		}
		lowest = low = 254; 	highest = high = 0;		i = index;
		mean_sum = mean = mode_sum = mode = median_sum = median = sum = 0;
		while( j ) {							// walk back from end gathering stats
			if( LOW || stutter_mon ) {
				low = fps_stats[ --i ];
				if( low < lowest ) lowest = low;
			}
			if( HIGH ) {
				high = fps_stats[ --i ];
				if( high > highest ) highest = high;
			}	
			if( MEAN )   mean_sum   += fps_stats[ --i ];
			if( MODE )   mode_sum   += fps_stats[ --i ];
			if( MEDIAN ) median_sum += fps_stats[ --i ];
			sum += fps_stats[ --i ];
			if( i <= 0 ) i = array_size;
			j -= DATA_SIZE;
			if( mid > 0 && len - j === mid ) {
				// both are reporting, store data for shorter one
				let m = 0;
				mid_data[ m++ ] = round( sum / mid_div ); // fps
				if( MEDIAN ) mid_data[ m++ ] = round( median_sum / mid_div );
				if( MODE )   mid_data[ m++ ] = round( mode_sum   / mid_div );
				if( MEAN )   mid_data[ m++ ] = round( mean_sum   / mid_div );
				if( HIGH )   mid_data[ m++ ] = highest;
				if( LOW || stutter_mon ) mid_data[ m++ ] = lowest;
			}
		}
		data[ i++ ] = round( sum / div ); 		// fps
		i = 0;
		if( MEDIAN ) data[ i++ ] = round( median_sum / div );
		if( MODE )   data[ i++ ] = round( mode_sum   / div );
		if( MEAN )   data[ i++ ] = round( mean_sum   / div );
		if( HIGH )   data[ i++ ] = highest;
		if( LOW || stutter_mon ) data[ i++ ] = lowest;
		do_callback();
		if( !short_rpt ) set_time_strs();			// just-in-time init'n
		if( short_rpt === 0 ) 
			rpt_results( short_data, short_term_str );
		if( long_rpt === 0 ) 
			rpt_results( long_data, long_term_str );
		if( index === array_size ) 
			_reset_fps_monitor( delta );			// memory full, reset counters
	}

	return {	
				  _setup_fps_calc: _setup_fps_calc,
			    _setup_fps_report: _setup_fps_report, 
				_init_fps_monitor: _init_fps_monitor,
			 _turn_on_fps_monitor: _turn_on_fps_monitor,
			_turn_off_fps_monitor: _turn_off_fps_monitor,
			   _reset_fps_monitor: _reset_fps_monitor,
					_set_callback: _set_callback,
				  _clear_callback: _clear_callback,
					_realtime_fps: _realtime_fps, 
					 _current_fps: _current_fps, 
				  _short_term_fps: _short_term_fps, 
				   _long_term_fps: _long_term_fps, 
					  _get_frames: _get_frames, 
				    _get_fps_data: _get_fps_data, 
					 _fps_monitor: _fps_monitor
			};
			
}).call(this);


